loadPrerequisites=function()
{
  require(arm)
  require(bayesglm)
}

getTestAndTrainInds=function(individuals, crossValDetails, returnNames=FALSE, DEBUG=1, verbose=1)
{
  crossValFold = crossValDetails[["fold"]]
  numTestSets  = crossValDetails[["numRepeats"]]
  numInds      = length(individuals)

  testInds  = as.list(rep(NA, numTestSets))
  trainInds = as.list(rep(NA, numTestSets))

  # Break up the individuals into a set of training and test sets, 
  # first decide how many individuals to have in each block (res holds these breaks)
  isNextInt=function(inVec, index) 
    return(floor(inVec[index+1])==(floor(inVec[index])+1))
  indBreaks = (1:numInds)/(numInds/crossValFold)
  res = sort(unique(c(0, which(sapply(1:(length(indBreaks)-1), FUN=isNextInt, inVec=indBreaks))+1, numInds)))
  if(crossValDetails[["balanceCaseCon"]])
  {
    whichCases    = which(individuals==1)
    whichControls = which(individuals==0)
    cases    = individuals[whichCases]
    controls = individuals[whichControls]
    tempCrossValDetails = crossValDetails
    tempCrossValDetails[["balanceCaseCon"]] = FALSE
    caseSets = getTestAndTrainInds(individuals=cases   , crossValDetails=tempCrossValDetails, returnNames=TRUE)
    contSets = getTestAndTrainInds(individuals=controls, crossValDetails=tempCrossValDetails, returnNames=TRUE)
    for(testSeti in 1:numTestSets)
    {
      # NOTE, that because individuals in my sim data are drawn from the 5000 WITH REPLACEMENT, can have non unique individual names.
      # This applis both to within the case/control groups, but also between cases and controls. Non ideal naming scheme, oops.
     #testInds[[ testSeti]] = match(c(caseSets[["testInds"]][[ testSeti]], contSets[["testInds"]][[ testSeti]]), names(individuals))
     #trainInds[[testSeti]] = match(c(caseSets[["trainInds"]][[testSeti]], contSets[["trainInds"]][[testSeti]]), names(individuals))
      testInds[[ testSeti]] = c(whichCases[   match(caseSets[["testInds" ]][[ testSeti]], names(cases)   )], 
                                whichControls[match(contSets[["testInds" ]][[ testSeti]], names(controls))])
      trainInds[[testSeti]] = c(whichCases[   match(caseSets[["trainInds"]][[ testSeti]], names(cases)   )], 
                                whichControls[match(contSets[["trainInds"]][[ testSeti]], names(controls))])
      propCasesChecktest  = sum(individuals[testInds[[ testSeti]]]==1)/(sum(individuals[testInds[[ testSeti]]]==1)+sum(individuals[testInds[[ testSeti]]]==0))
      propCasesChecktrain = sum(individuals[trainInds[[testSeti]]]==1)/(sum(individuals[trainInds[[testSeti]]]==1)+sum(individuals[trainInds[[testSeti]]]==0))
      if(abs(propCasesChecktest-propCasesChecktrain)                  >numTestSets |
         abs(propCasesChecktest-(length(whichCases)/sum(individuals)))>numTestSets ) {
        if(verbose>=1)
          cat("  Warning, supposed to be matching the proportion of cases and controls between test and train sets\n")
        if(DEBUG>=2)
          browser()
      }
    }
    rv = list("testInds"   = testInds    ,
              "trainInds"  = trainInds   )
    return(rv)
  }

  for(testSeti in 1:numTestSets)
  {
    if(crossValDetails$randomInds)
      randomisedCCSInds   = sample((1:numInds), numInds, replace=FALSE) else
      randomisedCCSInds   = (1:numInds)
    testInds[[ testSeti]] = sort(randomisedCCSInds[(res[testSeti]+1):res[testSeti+1]])
    trainInds[[testSeti]] = (1:numInds)[-testInds[[testSeti]]]
    if(returnNames)
    {
      testInds[[ testSeti]] = names(individuals)[testInds[[ testSeti]]]
      trainInds[[testSeti]] = names(individuals)[trainInds[[testSeti]]]
    }
  }
  rv = list("testInds"   = testInds    ,
            "trainInds"  = trainInds   )
  return(rv)
}


calcCrossValLikelihood=function(coefficients  ,  testSampleDat,  testSampleOutcomes,
                                lnullModelLike, trainSampleDat, trainSampleOutcomes,
                                simpleReturn=TRUE, DEBUG=1, verbose=DEBUG
                               #BFbase=exp(1)
                               )
{
  # Calculate answers:
  logitProbs = cbind(rep(1, nrow(testSampleDat)), testSampleDat)%*%coefficients
  probs      = expit(logitProbs) #, base=BFbase)
  lprob      = sum(log(probs)) #, base=BFbase))
  lthisSetModelLike = sum(dbinom(x=testSampleOutcomes, size=1, 
                          prob=probs, log=TRUE))
  lthisSetNullLike  = sum(dbinom(x=testSampleOutcomes, size=1, 
                          prob=mean(trainSampleOutcomes), log=TRUE))

  # DEBUGGING/TESTING CODE
  if(DEBUG>2 | verbose>2 | 0) {
    cat("  \n")
    cat("  lthisSetModelLike        = ", lthisSetModelLike        , "\n", sep="")
    cat("  lthisSetNullLike         = ", lthisSetNullLike         , "\n", sep="")
    cat("  loverallNullLike         = ", lnullModelLike           , "\n", sep="")
    cat("  mean(testCCS )           = ", mean( testSampleOutcomes), "\n", sep="")
    cat("  mean(trainCCS)           = ", mean(trainSampleOutcomes), "\n", sep="")
    trainSubsetRgenPoly = trainSampleDat[, colSums(trainSampleDat)>0,drop=FALSE]
    testSubsetRgenPoly  = testSampleDat[ , colSums(testSampleDat )>0,drop=FALSE]
    if(var(testSampleOutcomes)==0)
      cat("  testCCS are all of one type.\n") else
      cat("  max(cor_test( gen, ccs)) = ", max(cor( testSampleOutcomes,  testSubsetRgenPoly)), "\n", sep="")
    if(var(trainSampleOutcomes)==0)
      cat("  trainCCS are all of one type.\n") else
      cat("  max(cor_train(gen, ccs)) = ", max(cor(trainSampleOutcomes, trainSubsetRgenPoly)), "\n", sep="")
  }
 #if((abs(lthisSetNullLike-lthisSetModelLike)> 100 |
 #        lthisSetNullLike-lthisSetModelLike >(50/DEBUG+1) ) & DEBUG>0) {
 #  cat("  Warning, highly different likelihoods! (fitted=", 
 #      lprob, ", and null=", lthisSetNullLike, ") \n", sep="")
 #  browser()
 #}
  if(DEBUG>2)
    browser()

  # Return values:
  if(simpleReturn)
    return(lthisSetModelLike)    
  rv         = list("probs"            = probs,
                    "lthisSetNullLike" = lthisSetModelLike
                   )
  return(rv)
}

# This function is used to get the scale for the prior on the log(ODDS RATIO)
# The function expects to be given ODDS RATIOS as guidelines, which it then logs, convert if using relative risks
getScales=function(SNPdf=1, DEBUG=1, verbose=DEBUG, usage="bestAprioriGuess",
                   numSNPReg=NULL, pNumHits = 2, 
                   pMinHitSize=RRtoOR(1.2, prev=prev), 
                   prev=0.01, specifyMAFdepByCommon=TRUE, 
                   baseMAF = 0.25, MAFs=NULL, MAFscalePower=1/2, maxMAFmultiplier=25/0.5, maxScale=0.5)
{
  if(is.null(usage)) {
    cat("  Error, you must specify how to generate scale (specify 'usage', probably use bestAprioriGuess)\n")
    browser()
  }
  if(usage=="autoExplore")
    inScales = getDeltas() else
  if(usage=="bestAprioriGuess") {
    if(numSNPReg<=pNumHits) {
      if(DEBUG > 2 | verbose >= 2)
        cat("  Mild warning: Not enough SNPs for this prior to make sense (in getScales)! Setting scale to ", maxScale, ".\n")
      return(maxScale)
    }
    # If we believed that ~2 variants were causal we might implement this by constraining the Cauchy to have quantiles:
    # c("2/numSNPReg"= 1.1, "1-(2/numSNPReg)"= 1.1)
    # The 2's in the denom are because we split the quants between upper and lower
    if(verbose>3)
      cat("  Calculating quantiles in getScales...")
    quantiles = rbind(c(pNumHits/(2*numSNPReg), 1-(pNumHits/(2*numSNPReg))), 
                        c(log(pMinHitSize)       , log(1/pMinHitSize)     ))
    if(verbose>3)
      cat("  Done\n")
    dimnames(quantiles) = list(c("quantiles", "effectSizes"), c("lower", "upper"))
    # This approach to scaling currently assumes symmetry (I'm probably forced to, at least for now)
    if(SNPdf==1) {
      if(verbose>1.5)
        cat("  Using a Cauchy distribution.\n")

      if(verbose>3 | DEBUG>=3) {
        cat("  quantiles are:\n")
        print(quantiles)
        cat("  quantiles[\"quantiles\",1] is:", quantiles["quantiles",1], "\n")
        cat("  Calculating scale getScales (Cauchy distribution)...")
        if(DEBUG>=3)
          browser()
      }
      scale          = quantiles["effectSizes",1] /
                       qcauchy(    p=quantiles["quantiles",1], location = 0, scale = 1      , lower.tail = FALSE)
      if(verbose>3)
        cat("  Done\n")
     #quantileCheck  = exp(qcauchy(p=quantiles["quantiles", ], location = 0, scale = scaling, lower.tail = TRUE))
    } else if(SNPdf==Inf) {
      if(verbose>1.5)
        cat("  Using a Normal distribution.\n")
      if(verbose>3)
        cat("  Calculating scale getScales (Normal distribution)...")
        scale  = quantiles["effectSizes",1] /
                 qnorm(p=quantiles["quantiles",1], sd=1, lower.tail = FALSE)
      if(verbose>3)
        cat("  Done\n")
    } else {
      cat("  Warning, general t-distributions have not yet been implemented.\n")
      browser()
    }
  } else if(usage=="MAFdependent") {
    if(specifyMAFdepByCommon)
      baseScale = getScales(DEBUG=DEBUG, verbose=verbose  , usage="bestAprioriGuess",  
                            SNPdf=SNPdf, pNumHits=pNumHits, pMinHitSize=pMinHitSize , numSNPReg=numSNPReg)
      if(is.null(MAFs)) {
        cat("  Error, can't do MAF dependent scales without MAFs!\n")
        browser()
      }
      if((numSNPReg)!=length(MAFs)) {
        cat("  Error, MAFs should have the same length as the number of SNPs in the region!\n")
        browser()
      }

      if(SNPdf==1)
        fiftyPercentRR = ORtoRR(exp(qcauchy(p=0.25, scale=baseScale, lower.tail = FALSE)), prev=prev) else
        fiftyPercentRR = ORtoRR(exp(qnorm(  p=0.25, sd   =baseScale, lower.tail = FALSE)), prev=prev) # Note that scale maps to sd, NOT variance
      # The following makes sense as: A Cauchy is defined so that 50% of the observations are more extreme than 'scale'
      mafAdjustedRRs = 1+(fiftyPercentRR-1)*pmin(((baseMAF/MAFs)^(MAFscalePower)), maxMAFmultiplier)

      allScales = log(RRtoOR(mafAdjustedRRs, prev=prev))
     #allScales = baseScale*pmin(((baseMAF/MAFs)^(MAFscalePower)), maxMAFmultiplier)
      if(max(allScales)>maxScale) {
        extremeLogORs = qcauchy(p=c(0.1, 0.05), location = 0, scale = max(allScales), lower.tail = FALSE)
        if((verbose>=2 & DEBUG>=2) | DEBUG>2) {
          cat("  max(scales) = ", max(allScales), ", which means that log odds ratios greater than ",  
              round(extremeLogORs[1], 1), "/", round(extremeLogORs[2], 3), 
              " \n  would have probability 10%/5% under the prior at this SNP.\n", sep="")
          scaleSummary = matrix(c(MAFs*100, allScales, ORtoRR(exp(allScales), prev=0.01)), nrow=3, byrow=TRUE)
          rownames(scaleSummary) = c("MAF", "logORscale", "converting to RR")
          print(round(scaleSummary[, 1:min(10, ncol(scaleSummary))], 3))
          # x=rcauchy(1e4); x=x[abs(x)<10]; propBigger=function(vec, cutoff) return(sum(abs(vec)>cutoff)/length(vec)); 
          # cutOffs=c(1:5); sapply(cutOffs, propBigger, vec=x)
          # What the above commented line shows is that >25% of the prior is more extreme than _twice_ the log(ORs) reported by scale
          # plot(density(x)); abline(v=c(-1,1), col="grey"); 
          if(DEBUG>2)
            browser()
        }
        allScales = pmin(allScales, maxScale)
      }
      if(DEBUG>3) {
        cat("  Are the scales (MAF dependent) the ones you want?\n")
        browser()
      }
      return(allScales)
  } else {
    cat("  Only autoExplore and bestAprioriGuess implemented at present.\n")
    browser()
  }
  if(verbose>3)
    cat("  Returning scale.\n")  
  return(min(maxScale, abs(scale)))
  
  
}
# install.packages("mnormt", lib=.libPaths()[1], contriburl="http://cran.cnr.Berkeley.edu")

# Note that only small and medium Cauchy performed well except: small regions with rare-only and some marginally goot 'normal' results
# getRegionAssocModels()
getRegionAssocModels=function(numSNPReg    = "unknown"               , # If scale depends on region size, need this for scales here
                              useNumSNPreg = TRUE                    , # Should the scale depend on the number of SNPs in the region?
                              modelNames   = c("Cs", "Cm", "Cl", "Ns", "Nl", "CsMAF", "ClMAF"),
                              modelsToUse  = c(1  , 1  , 1  , 1  , 1  , 1  , 1  )   , # 0=don't use this model, 1=use this model
                              pNumHits     = c(2  , 4  , 8  , 3  , 6  , 2  , 6  )   , # Under alt. roughly how many causal variants?
                              pMinHitSize  = c(1.2, 1.5, 3.0, 1.5, 2.0, 1.1, 1.3)   , # Specified in relative risk space
                              SNPdf        = c(1  , 1  , 1  , Inf, Inf, 1  , 1  )   , # 1=Cauchy, Inf = Normal
                              MAFdependES  = c(0  , 0  , 0  , 0  , 0  , 1  , 1  )   , # Should the effect size depend on MAF (0=no, 1=yes)
                              biggerESatNS = c(0  , 0  , 0  , 0  , 0  , 0  , 0  )   , # Should the effect sizes be bigger for NS SNPs? (0=no, 1=yes) 
                             #modelWeights = c(1  , 1  , 1  , 0.5, 0.5, 1.5, 1.5)   , # Prior weight on each model
                              modelWeights = c(1  , 1  , 1  , 1  , 1  , 1  , 1  )   , # Prior weight on each model
                             #weightModels = TRUE                    , # Should models be weighted (removed as this is for elsewhere)
                              prev         = 0.01                      # Required to convert OR<-->RR
                             )
{
  scales = as.list(rep(NA, length(pNumHits)))
  for(scaleIndex in 1:length(scales))
  {
    # At present I think I calculate this within the inference function. 
    # Here would probably be better, but I don't know the number of SNPs in the region?
    if(is.numeric(numSNPReg))
      scales[[scaleIndex]] = getScales(SNPdf       = SNPdf, verbose=DEBUG, usage="bestAprioriGuess",  
                               numSNPReg   = ncol(rGen), 
                               pNumHits    = pNumHits, 
                               pMinHitSize = RRtoOR(OR=pMinHitSize, prev=prev))
  }
  longModelNames = rep(NA, length(modelsToUse))
  for(mni in 1:length(modelsToUse))
    longModelNames[mni] = paste("SNPdf="     , SNPdf      [mni], "_", 
                                "pNumHits"   , pNumHits   [mni], "_", 
                                "pMinHitSize", pMinHitSize[mni], "_", 
                                "MAFdependES", MAFdependES[mni], "", 
                                sep="")  
  models = list("pNumHits"       = pNumHits      ,
                "SNPdf"          = SNPdf         ,
                "longModelNames" = longModelNames,
                "modelNames"     = modelNames    ,
                "pMinHitSize"    = pMinHitSize   ,
                "pMinHitSize"    = pMinHitSize   ,
                "modelWeights"   = modelWeights  ,
                "MAFdependES"    = MAFdependES   ,
                "scales"         = scales        , 
                "biggerESatNS"   = biggerESatNS  ,
                "modelsToUse"    = modelsToUse   ,
                "numModels"      = sum(modelsToUse),
                "prev"           = prev
               #"weightModels"   = weightModels
               )
  return(models)
}

getCrossValDetails=function(numRepeats=NA, numRepeatsOfCrossVal=20, fold=5)
{
  rv = list("numRepeatsOfCrossVal" = numRepeatsOfCrossVal, # Number of times to calculate the likelihood of the (1/fold) test individuals numRepeats times.
            "numRepeats"           = numRepeats,           # Number of fit/test partitions to produce (should be equal to fold)
            "fold"                 = fold      ,           # 1/fold is the proportion of individuals placed in the test set
            "replace"              = FALSE     ,           # Only 'FALSE' implemented. fit/test individuals not drawn with replacement. 
            "randomInds"           = TRUE      ,           # Use randomised partitions (rather than sequential) for fit/test sets
            "balanceCaseCon"       = FALSE     )           # Condittion on the same proportions (as near as poss) of cases and controls in each partition?
  if(!rv$replace)
    rv$numRepeats = rv$fold
  return(rv)
}

getModelChoiceDetails=function()
{
  rv = list("modelAverage" = TRUE,
            "weightModels" = TRUE)
  return(rv)
}

getAnalysisDetails=function(incWeakManyRare=TRUE)
{
  priors       = c("cauchy"="small", "cauchy"="medium", "cauchy"="large", 
                   "normal"="small")
  parStrengths = c("onlyRare"    ,
                   "weakFixedManyRare",
                   "noCommon"    ,
                   "sMultiCommon",
                   "smaller"     ,
                   "mostlyCommon",
                   "noRare"      ,
                   "onlyCommon"  ,
                   "singleCommon",
                   "nullData"
                  )
  if(!incWeakManyRare)
    parStrengths=parStrengths[!parStrengths=="weakFixedManyRare"]
  priorDetails = list("small" =c("pNumHits"=2, "numSNPReg"=NA, "pMinHitSize"=1.2),
                      "medium"=c("pNumHits"=4, "numSNPReg"=NA, "pMinHitSize"=1.3),
                      "large" =c("pNumHits"=8, "numSNPReg"=NA, "pMinHitSize"=2.0)
                     )
  genes        = c(20, 14, 4 , 16, 25,
                   11, 5 , 13, 6 , 24, 
                   23, 18, 21, 19, 15) #, 17, 8, 9, 12, 22, 26, 7, 10 #these commented all have >1500 SNPs
  nonGenicRegions   = c(1:23) #, 24:47) #24:26 have 1001SNPs, 27 has 2001
  numPermutes       = 1000
  moreGenicPermutes = 1
  rv = list("priors"            = priors           ,
            "parStrengths"      = parStrengths     ,
            "priorDetails"      = priorDetails     ,
            "genes"             = genes            ,
            "nonGenicRegions"   = nonGenicRegions  ,
            "numPermutes"       = numPermutes      ,
            "moreGenicPermutes" = moreGenicPermutes)
}


